require "digest"
require "delegate"
require 'net/scp'

module BundleDepot
  class BundleDepotError     < StandardError; end
  class GemfileNotFound      < BundleDepotError; end
  class RemoteConfigurationMissing < BundleDepotError; end
  class BundleNotFound       < BundleDepotError; end
  class RemoteBundleNotFound < BundleNotFound; end
  class LocalBundleNotFound  < BundleNotFound; end
  class ArchivingFailed      < BundleDepotError; end

  class Store
    def cached?(file)
      raise NotImplementedError
    end

    def store(file)
      raise NotImplementedError
    end

    def fetch(file, dest_dir)
      raise NotImplementedError
    end

    def with_packing
      PackingDecorator.new(self)
    end
  end

  class SCPStore < Store
    attr_reader :host, :user, :password, :path
    
    def initialize(host, user, password, path = ".")
      @host     = host
      @user     = user
      @password = password
      @path     = path
    end

    def cached?(file)
      target = File.join(path, File.basename(file))
      output = "false"

      Net::SSH.start(host, user, { password: password}) do |session|
        output = session.exec! "test -e #{target} && echo 'true'"
      end

      output && output.strip == "true" 
    end

    def store(file)
      Net::SSH.start(host, user, { password: password}) do |session|
        session.exec! "mkdir -p #{path}"
        session.scp.upload! file, File.join(path, File.basename(file))
      end
    end

    def fetch(file, dest_dir)
      source = File.join(path, file)  
      target = File.join(dest_dir, file) 
      Net::SCP.download!(host, user, source, target, ssh: { password: password })
    rescue Net::SCP::Error
      raise BundleNotFound
    end

  end

  class FileSystemStore < Store
    attr_reader :path

    def cached?(file)
      File.exists? File.join(path, file)
    end

    def initialize(path)
      @path = path
      FileUtils.mkdir_p(path)
    end

    def store(file)
      FileUtils.cp_r file, File.join(path, File.basename(file))
    end

    def fetch(file, dest_dir)
      FileUtils.cp_r File.join(path, file), dest_dir
    end
  end

  class PackingDecorator < SimpleDelegator

    def cached?(file)
      super(packed_file(file))
    end

    def store(file)
      working_dir = File.expand_path(File.dirname(file))
      source      = File.basename(file)
      target      = packed_file(file)

      raise BundleNotFound unless File.exists? file

      unless File.exist?(target)
        puts "=> Packing bundle"
        `tar -C #{working_dir} -czf #{target} #{source}`
        raise ArchivingFailed unless $?.exitstatus == 0
      end  

      super(target)
    end

    def fetch(file, dest_dir)
      source = packed_file(file)
      target = File.join(dest_dir, source)
      
      super(source, dest_dir)
      
      puts "=> Unpacking #{target}"
      `tar -C #{dest_dir} -xf #{target}`
      raise ArchivingFailed unless $?.exitstatus == 0
    end

    def packed_file(file)
      file + ".tar.gz"
    end
  end

  class Cache
    DEFAULT_BUNDLE_DEPOT_CURRENT = File.join(".bundle", "depot", "current")
    DEFAULT_BUNDLE_DEPOT_CACHE   = File.join(".bundle", "depot", "cache")

    attr_reader :local, :remote

    def initialize(local  = FileSystemStore.new(Cache.cache_path), 
                   remote = FileSystemStore.new(File.join(".bundle", "depot", "remote")).with_packing)

      if remote_configured? && !ENV["TEST"]
        remote = SCPStore.new(ENV["BUNDLE_DEPOT_SCP_HOST"], 
                             ENV["BUNDLE_DEPOT_SCP_USER"],
                             ENV["BUNDLE_DEPOT_SCP_PASS"], "cache").with_packing
      end

      @local  = local
      @remote = remote
    end

    def store
      setup
      store_remote
    end
    
    def fetch
      setup
      fetch_remote unless fetch_local
      manage_symlink
    end

    def fetch_local 
      if local.cached?(digest)
        puts "=> Bundle found in local cache"
        true
      else
        puts "=> Bundle is not in local cache"
      end
    end

    def fetch_remote
      return unless remote_configured?

      if remote.cached?(digest)
        puts "=> Bundle found in remote cache. Downloading!"
        remote.fetch(digest, local.path)
      else
        puts "=> Bundle is not in remote cache"
      end
    end

    def store_remote
      raise RemoteConfigurationMissing unless remote_configured?

      if remote.cached?(digest)
        puts "=> Bundle is already in remote cache. Skipping upload."
      else
        puts "=> Uploading bundle to remote cache"
        remote.store(File.join(local.path, digest))
      end
    end

    def self.cache_path
      ENV["BUNDLE_DEPOT_CACHE"] || DEFAULT_BUNDLE_DEPOT_CACHE
    end

    def self.current_bundle_path
      ENV["DEFAULT_BUNDLE_DEPOT_CURRENT"] || DEFAULT_BUNDLE_DEPOT_CURRENT
    end

    def remote_configured?
      ENV["BUNDLE_DEPOT_SCP_HOST"]
    end

    private

    def setup
      FileUtils.mkdir_p(File.dirname(Cache.current_bundle_path))
    end

    def manage_symlink
      source = File.join(local.path, digest)
      target = Cache.current_bundle_path

      FileUtils.mkdir_p(source) unless File.exists?(source)
      
      if remove_outdated_symlink(source, target)
        puts "=> Linking #{target} -> #{source}"
        File.symlink File.expand_path(source), target 
      else
        puts "=> #{target} is pointing to the right bundle"
      end
    end

    def remove_outdated_symlink(source, target)
      return true unless File.exist?(target)
      raise "#{target} is not a symlink. Aborting to not mess stuff up..." unless File.symlink?(target)
      return false if File.symlink?(target) && File.readlink(File.expand_path(target)) == File.expand_path(source)

      File.unlink(target)
    end

    def digest 
      raise GemfileNotFound unless File.exist?("Gemfile.lock")
      Digest::SHA2.file("Gemfile.lock").hexdigest
    end
  end

end
